package pp;

import java.util.Random;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;

import static pp.PhilosopherExperiment.MAX_EATING_DURATION_MS;
import static pp.PhilosopherExperiment.MAX_THINKING_DURATION_MS;

public class Philosopher extends Thread implements IPhilosopher {
	private final Random random;
	private volatile boolean stop;

	private int seat;
	private Philosopher left;
	private Philosopher right;
	private Lock table;
	public boolean eating;
	public Condition condition;


	public Philosopher() {
		this.random = new Random();
		this.stop = false;
		this.eating = false;
	}

	@Override
	public void run() {
		log("starting");
		try {
			while (!this.stop) {
				think();
				eat();
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		}
		log("stopped;");
	}

	private void think() throws InterruptedException {
		Thread.sleep(this.random.nextInt(MAX_THINKING_DURATION_MS));
	}

	private void eat() throws InterruptedException {
		this.table.lock();
		try {
			while (this.left.eating || this.right.eating) {
				this.condition.await();
			}
			log("left nor right eating");
			this.eating = true;
			log("eating");
		} finally {
			table.unlock();
		}
		Thread.sleep(this.random.nextInt(MAX_EATING_DURATION_MS));
		this.table.lock();
		try {
			this.eating = false;
			this.left.condition.signal();
			this.right.condition.signal();

			log("done eating");
		} finally {
			table.unlock();
		}
	}

	@Override
	public void stopPhilosopher() {
		log("stopping");
		this.stop = true;
		interrupt();
	}

	@Override
	public void setLeft(IPhilosopher left) {
		this.left = (Philosopher) left;
	}

	@Override
	public void setRight(IPhilosopher right) {
		this.right = (Philosopher) right;
	}

	@Override
	public void setSeat(int seat) {
		this.seat = seat;
	}

	@Override
	public void setTable(Lock table) {
		this.table = table;
		this.condition = this.table.newCondition();
	}

	private void log(String message) {
		synchronized (Philosopher.class) {
			for (var i = 1; i <= this.seat; i++) {
				System.out.print("                           ");
			}
			System.out.println(Thread.currentThread().getName()  + " " + message);
		}
	}
}
